/*
 * ext_io.c - Communication infrastructure for the external interface
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 Network Robots
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "ext.h"


static FILE *stream = NULL;	/* NULL if "ext" subsystem is not active */
static pid_t pid = 0;		/* 0 if there is no process */
static char in_path[100]; /* @@@ */
static char out_path[100]; /* @@@ */
static char **my_argv;		/* undefined if pid != 0 */


static void copy_argv(char **argv)
{
    int argc,i;

    for (argc = 0; argv[argc]; argc++);
    my_argv = malloc(sizeof(char *)*(argc+1));
    for (i = 0; i < argc; i++)
	my_argv[i] = stralloc(argv[i]);
    my_argv[i] = NULL;
}


static void free_argv(void)
{
    int i;

    for (i = 0; my_argv[i]; i++)
	free(my_argv[i]);
    free(my_argv);
}


static void child(char **argv)
{
    int fd;

    (void) close(0);
    fd = open(in_path,O_RDONLY);
    if (fd < 0) {
	perror(in_path);
	exit(1);
    }
    if (dup2(fd,0) < 0) {
	perror("dup2");
	exit(1);
    }
    fd = creat(out_path,0644);
    if (fd < 0) {
	perror(out_path);
	exit(1);
    }
    if (dup2(fd,1) < 0) {
	perror("dup2");
	exit(1);
    }
    if (execvp(*argv,argv) < 0) {
	perror(*argv);
	exit(1);
    }
    /* not reached */
}


static void cleanup_external(void)
{
    if (stream) ext_close();
}


FILE *ext_open(char **argv)
{
    static int need_exit_handler = 1;

    assert(!stream);
    sprintf(in_path,"__ext_%d.in",getpid());
    sprintf(out_path,"__ext_%d.out",getpid());
    stream = fopen(in_path,"w");
    if (!stream) {
	perror(in_path);
	exit(1);
    }
    if (need_exit_handler) {
	need_exit_handler = 0;
	atexit(cleanup_external);
    }
    copy_argv(argv);
    return stream;
}


int ext_read(void)
{
    int status;

    assert(stream);
    (void) fflush(stream);
    if (ferror(stream)) error("error writing to sub-process");
    if (!freopen(DEV_NULL,"r",stream)) { /* close old stream */
	perror("/dev/null");
	exit(1);
    }
    fflush(stdout);
    fflush(stderr);
    pid = fork();
    if (!pid) child(my_argv);
    if (waitpid(pid,&status,0) < 0) {
	perror("waitpid");
	exit(1);
    }
    if (!freopen(out_path,"r",stream)) {
	perror(out_path);
	exit(1);
    }
    return WIFEXITED(status) ? WEXITSTATUS(status) : -1;
}


void ext_close(void)
{
    assert(stream);
    if (fclose(stream) == EOF) {
	perror("fclose");
	exit(1);
    }
    stream = NULL;
    if (unlink(in_path) < 0) {
	perror(in_path);
	exit(1);
    }
    free_argv();
    if (pid) {
	if (unlink(out_path) < 0) {
	    perror(out_path);
	    exit(1);
	}
	pid = 0;
    }
}
